/* mktifs.c
 * Makes a TiFS file system on a disc.
 * (C) 1996 David Given
 */

#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>
#include <string.h>
#include <linux/fs.h>
#include "tifs.h"

#define BLOCKS(size) \
	(((size) / BLOCK_SIZE) + 1)

void help(void);

static struct option options[] =
{
	{"blocks",	1,	NULL,	'b'},
	{"bsize",	1,	NULL,	's'},
	{"root",	1,	NULL,	'r'},
	{"device",	1,	NULL,	'd'},
	{"help",	0,	NULL,	'?'},
	{NULL,		0,	NULL,	0},
};

int numblocks = 0;
int blocksize = 1024;
int rootentries = 256;

char *device = NULL;
int superblocksize;
TIFS_Superblock *sb;

int main(int argc, char **argv)
{
	int i;
	FILE *f;

	while(1)
	{
		int index = 0;

		i = getopt_long(argc, argv,
			"?r:b:s:d:", options, &index);
		if (i == -1)
			break;

		switch (i)
		{
			case 'b':	numblocks = atoi(optarg);
					break;

			case 's':	blocksize = atoi(optarg);
					break;

			case 'r':	rootentries = atoi(optarg);
					break;

			case 'd':	device = strdup(optarg);
					break;

			case '?':	help();
					exit(0);
					break;

			default:	help();
					exit(1);
		}
	}

	if (numblocks == 0)
	{
		fprintf(stderr, "You must specify the number of blocks\n");
		exit(1);
	}

	if (device == NULL)
	{
		fprintf(stderr, "You must specify the device\n");
		exit(1);
	}

	if (blocksize != 1024)
	{
		fprintf(stderr, "Only 1024 byte blocks are currently supported\n");
		exit(1);
	}

	numblocks--;

	superblocksize = sizeof(TIFS_Fileinfo)*rootentries +
		sizeof(TIFS_Superblock);

	{
		int ts = BLOCKS(superblocksize)*blocksize;

		rootentries = (ts - sizeof(TIFS_Superblock)) / sizeof(TIFS_Fileinfo);
		printf("Rounding up to %d entries\n", rootentries);
	}

	superblocksize = sizeof(TIFS_Fileinfo)*rootentries +
		sizeof(TIFS_Superblock);

	printf("Directory size = %d * %d + %d = %d B = %d blocks\n",
		sizeof(TIFS_Fileinfo), rootentries, sizeof(TIFS_Superblock),
		superblocksize,
		BLOCKS(superblocksize));

	
	if (superblocksize>0xFFFF)
	{
		fprintf(stderr, "Superblock must be less than 64kB\n");
		exit(1);
	}

	sb = (TIFS_Superblock*)malloc(superblocksize);
	sb->magic[0] = 'T';
	sb->magic[1] = 'i';
	sb->magic[2] = 'F';
	sb->magic[3] = 'S';
	sb->discsize = numblocks;
	sb->blocksize = blocksize;
	sb->firstusable = BLOCKS(superblocksize)+2;
	sb->boot_start = 0;
	sb->boot_size = 0;
	sb->numentries = rootentries;
	sb->numfilesused = 2;	/* Root and ino 0 */
	sb->numblocksused = sb->firstusable;

	strcpy(sb->file[TIFS_ROOT_INO].filename, " (root) ");
	sb->file[TIFS_ROOT_INO].filetype = TIFS_DIR;
	sb->file[TIFS_ROOT_INO].dirid = TIFS_ROOT_INO;
	sb->file[TIFS_ROOT_INO].mode = 0755;
	sb->file[TIFS_ROOT_INO].uid = 0;
	sb->file[TIFS_ROOT_INO].gid = 0;
	
	for(i=2; i<rootentries; i++)
		sb->file[0].filetype = TIFS_EMPTY;

	f = fopen(device, "r+");
	fseek(f, BLOCK_SIZE, SEEK_SET);
	fwrite(sb, superblocksize, 1, f);
	fclose(f);

	printf("Total file system size = %d * %d = %d kB\n",
		numblocks, blocksize, numblocks * blocksize / 1024);

	printf("Usable space = %d blocks = %d kB\n", 
		numblocks - sb->numblocksused,
		(numblocks - sb->numblocksused)*blocksize/1024);

	exit(0);
}

void help(void)
{
	printf("mktifs [options]\n");
	printf("    -d --device       Device to create file system on\n");
	printf("    -b --blocks       Size of filesystem in blocks\n");
	printf("    -s --bsize        Block size in bytes\n");
	printf("    -r --root         Size of root directory in files\n");
	printf("    -h --help         Displays this message\n");
}

