/* tifs.c
 * TiFS kernel module
 * (C) 1996 David Given
 */

#include <linux/module.h>
#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/module.h>
#include <linux/locks.h>
#include <linux/version.h>
#include <linux/malloc.h>
#include "tifs.h"

/* --- Configuration ------------------------------------------------------- */

#define CFG_ASSERT_ONENTRY	0  /* Set to check come entry values */
#define CFG_ASSERT_VERBOSE	1  /* If above set, print where error occured */
#define CFG_DEBUG_MSG		1  /* Print debug messages */
#define CFG_TRACE		1  /* If above set, print trace messages */

/* --- Forward declarations ------------------------------------------------ */

static struct super_block *tifs_read_super(struct super_block *, void *, int);

void tifs_read_inode(struct inode *);
void tifs_write_inode(struct inode *);
void tifs_close_inode(struct inode *);
void tifs_close_super(struct super_block *);
void tifs_write_super(struct super_block *);
void tifs_statfs(struct super_block *, struct statfs *, int);

int tifs_readdir(struct inode *, struct file *, void *, filldir_t);
int tifs_create(struct inode *, const char *, int, int, struct inode **);
int tifs_lookup(struct inode *, const char *, int, struct inode **);
int tifs_unlink(struct inode *, const char *, int);
int tifs_mkdir(struct inode *, const char *, int, int);
int tifs_rmdir(struct inode *, const char *, int);
int tifs_rename(struct inode *, const char *, int,
		struct inode *, const char *, int, int);
int tifs_mknod(struct inode *, const char *, int, int, int);

int tifs_bmap(struct inode *, int);
void tifs_truncate(struct inode *);
int tifs_file_write(struct inode *, struct file *, const char *, int count);

int tifs_symlink(struct inode *, const char *, int, const char *);
int tifs_readlink(struct inode *, char *, int);
int tifs_followlink(struct inode *, struct inode *, int, int,
	struct inode **rinode);

/* --- Globals ------------------------------------------------------------- */

static struct file_system_type fs_block =
{
	tifs_read_super,	/* Entry point */
	"tifs",			/* FS name */
	1,			/* Needs device */
	NULL			/* Reserved */
};

static struct super_operations sops_block =
{
	tifs_read_inode,	/* Read inode */
	NULL,			/* Change notification */
	tifs_write_inode,	/* Write inode */
	tifs_close_inode,	/* Finished with inode */
	tifs_close_super,	/* Finished with superblock */
	tifs_write_super,	/* Write superblock */
	tifs_statfs,		/* Stat filing system */
	NULL			/* Remount */
};

/* Default fops and iops blocks (regular files) */

static struct file_operations fops_block =
{
	NULL,			/* lseek */
	generic_file_read,	/* read */
	tifs_file_write,	/* write */
	NULL,			/* readdir (bad) */
	NULL,			/* select */
	NULL,			/* ioctl */
	generic_file_mmap,	/* mmap */
	NULL,			/* open */
	NULL,			/* release */
	file_fsync,		/* fsync */
};

static struct inode_operations iops_block =
{
	&fops_block,		/* Default file operations */

	NULL,			/* create */
	NULL,			/* lookup */
	NULL,			/* link */
	NULL,			/* unlink */
	NULL,			/* symlink */
	NULL,			/* mkdir */
	NULL,			/* rmdir */
	NULL,			/* mknod */
	NULL,			/* rename */
	NULL,			/* readlink */
	NULL,			/* follow_link */
	generic_readpage,	/* readpage */
	NULL,			/* writepage */
	tifs_bmap,		/* bmap */
	tifs_truncate,		/* truncate */
	NULL,			/* permission */
	NULL			/* smap */
};

/* Sym-link inode operations */

static struct inode_operations iops_symlink_block =
{
	NULL,			/* No file operations */

	NULL,			/* create */
	NULL,			/* lookup */
	NULL,			/* link */
	NULL,			/* unlink */
	NULL,			/* symlink */
	NULL,			/* mkdir */
	NULL,			/* rmdir */
	NULL,			/* mknod */
	NULL,			/* rename */
	tifs_readlink,		/* readlink */
	tifs_followlink,	/* follow_link */
	NULL,			/* readpage */
	NULL,			/* writepage */
	NULL,			/* bmap */
	NULL,			/* truncate */
	NULL,			/* permission */
	NULL			/* smap */
};

/* Directory fops and iops blocks */

static struct file_operations fops_dir_block =
{
	NULL,			/* lseek */
	NULL,			/* read */
	NULL,			/* write (bad) */
	tifs_readdir,		/* readdir */
	NULL,			/* select */
	NULL,			/* ioctl */
	NULL,			/* mmap */
	NULL,			/* open */
	NULL,			/* release */
	file_fsync,		/* fsync */
};

static struct inode_operations iops_dir_block =
{
	&fops_dir_block,	/* Default file operations */

	tifs_create,		/* create */
	tifs_lookup,		/* lookup */
	NULL,			/* link */
	tifs_unlink,		/* unlink */
	tifs_symlink,		/* symlink */
	tifs_mkdir,		/* mkdir */
	tifs_rmdir,		/* rmdir */
	tifs_mknod,		/* mknod */
	tifs_rename,		/* rename */
	NULL,			/* readlink */
	NULL,			/* follow_link */
	NULL,			/* readpage */
	NULL,			/* writepage */
	NULL,			/* bmap */
	NULL,			/* truncate */
	NULL,			/* permission */
	NULL			/* smap */
};

/* --- Utility macros and functions ---------------------------------------- */

/* Debug messages */

#if CFG_DEBUG_MSG == 1
#define DEBUG(x) printk x
#if CFG_TRACE == 1
#define TRACE(x) printk("tifs:" x "\n")
#else
#define TRACE(x) /* */
#endif
#else
#define DEBUG(x) /* */
#define TRACE(x) /* */
#endif

/* Gets our workspace from an inode. */

#if ASSERT_ONENTRY == 1
#if ASSERT_VERBOSE == 1
#define NULLERROR(what, name) \
	"tifs: null " what " in " name "!\n"
#else
static char assertionfailure[] = "tifs: Assertion failure!\n";
#define NULLERROR(what, name) \
	assertionfailure
#endif

#define INIT_SB_TS(name, inode) \
	struct super_block *sb;						\
	TIFS_Superblock *ts;						\
	if (!inode)							\
	{								\
		iput(inode);						\
		printk(NULLERROR("inode", name));			\
		return -EBADF;						\
	}								\
	sb = inode->i_sb;						\
	ts = (TIFS_Superblock *)sb->u.generic_sbp;			\
	if (!ts)							\
	{								\
		iput(inode);						\
		printk(NULLERROR("ts", name));				\
		return -EBADF;						\
	}
#else
#define INIT_SB_TS(name, inode) \
	struct super_block *sb = inode->i_sb;				\
	TIFS_Superblock *ts = (TIFS_Superblock *)sb->u.generic_sbp
#endif

/* A simple semaphore. */

#define LOCKSB \
	lock_super(sb)

#define UNLOCKSB \
	unlock_super(sb)

/* Converts a size to a number of blocks */

#define BLOCKS(size) \
	(((size) / BLOCK_SIZE) + 1)

/* Setting and retrieving superblock values */

#define GET_SB_MAGIC(c)		ts->magic[c]
#define GET_SB_BLOCKSIZE	ts->blocksize
#define GET_SB_DISCSIZE		ts->discsize
#define GET_SB_FIRSTUSABLE	ts->firstusable
#define GET_SB_BOOTSTART	ts->bootstart
#define GET_SB_BOOTSIZE		ts->bootsize
#define GET_SB_NUMENTRIES	ts->numentries
#define GET_SB_NUMFILESUSED	ts->numfilesused
#define GET_SB_NUMBLOCKSUSED	ts->numblocksused

#define SET_SB_MAGIC(c, v)	ts->magic[c] = v
#define SET_SB_BLOCKSIZE(v)	ts->blocksize = v
#define SET_SB_DISCSIZE(v)	ts->discsize = v
#define SET_SB_FIRSTUSABLE(v)	ts->firstusable = v
#define SET_SB_BOOTSTART(v)	ts->bootstart = v
#define SET_SB_BOOTSIZE(v)	ts->bootsize = v
#define SET_SB_NUMENTRIES(v)	ts->numentries = v
#define SET_SB_NUMFILESUSED(v)	ts->numfilesused = v
#define SET_SB_NUMBLOCKSUSED(v)	ts->numblocksused = v

/* Setting and retrieving file entry values */

#define GET_FE_FILENAME(f, buf)	\
	strncpy(buf, ts->file[f].filename, TIFS_NAMELENGTH)
#define GET_FE_FILETYPE(f)	ts->file[f].filetype
#define GET_FE_DIRID(f)		ts->file[f].dirid
#define GET_FE_START(f)		ts->file[f].start
#define GET_FE_BLOCKS(f)	ts->file[f].blocks
#define GET_FE_MODE(f)		ts->file[f].mode
#define GET_FE_UID(f)		ts->file[f].uid
#define GET_FE_GID(f)		ts->file[f].gid
#define GET_FE_SIZE(f)		ts->file[f].size
#define GET_FE_ATIME(f)		ts->file[f].atime
#define GET_FE_MTIME(f)		ts->file[f].mtime
#define GET_FE_CTIME(f)		ts->file[f].ctime

#define SET_FE_FILENAME(f, buf, len) \
	do {								\
		strncpy(ts->file[f].filename, buf, len);		\
		ts->file[f].filename[len] = '\0';			\
	} while(0)
#define SET_FE_FILETYPE(f, v)	ts->file[f].filetype = v
#define SET_FE_DIRID(f, v)	ts->file[f].dirid = v
#define SET_FE_START(f, v)	ts->file[f].start = v
#define SET_FE_BLOCKS(f, v)	ts->file[f].blocks = v
#define SET_FE_MODE(f, v)	ts->file[f].mode = v
#define SET_FE_UID(f, v)	ts->file[f].uid = v
#define SET_FE_GID(f, v)	ts->file[f].gid = v
#define SET_FE_SIZE(f, v)	ts->file[f].size = v
#define SET_FE_ATIME(f, v)	ts->file[f].atime = v
#define SET_FE_MTIME(f, v)	ts->file[f].mtime = v
#define SET_FE_CTIME(f, v)	ts->file[f].ctime = v

/* --- Module stuff -------------------------------------------------------- */

int init_module(void)
{
	int err;
	
	printk("TiFS V0.1.%d (C) 1997 David Given\n", COMPILEVERSION);

	err = register_filesystem(&fs_block);

	if (err)
	{
		printk("tifs: failed to register filesystem\n");
		return err;
	}

	return 0;
}

void cleanup_module(void)
{
	int err;
	if (MOD_IN_USE)
		printk("tifs: filesystem in use, removal delayed\n");

	DEBUG(("tifs: unregistering\n"));

	err = unregister_filesystem(&fs_block);
	if (err)
	{
		printk("tifs: failed to unregister filesystem\n");
		return;
	}
}

/* --- Superblock ---------------------------------------------------------- */

struct super_block *tifs_read_super(struct super_block *sb, void *data,
	int silent)
{
	struct buffer_head *bh;
	TIFS_Superblock *ts;
	kdev_t dev = sb->s_dev;
	int superblocksize;
	int blocksize;
	int i;

	MOD_INC_USE_COUNT;

	/* Load the superblock header (but not the file info, we don't know
	   how much there is yet */

	TRACE("loading basic superblock");

	set_blocksize(dev, BLOCK_SIZE);
	LOCKSB;
	bh = bread(dev, 1, BLOCK_SIZE);

	if (!bh)
	{
		sb->s_dev = 0;
		UNLOCKSB;
		printk("tifs: unable to read superblock\n");
		MOD_DEC_USE_COUNT;
		return NULL;
	}

	ts = (TIFS_Superblock *)bh->b_data;

	/* Now it's loaded, we check the magic number. */

	if ((GET_SB_MAGIC(0) != 'T') ||
	    (GET_SB_MAGIC(1) != 'i') ||
	    (GET_SB_MAGIC(2) != 'F') ||
	    (GET_SB_MAGIC(3) != 'S'))
	{
		sb->s_dev = 0;
		UNLOCKSB;
		brelse(bh);
		printk("tifs: bad magic number\n");
		MOD_DEC_USE_COUNT;
		return NULL;
	}

	blocksize = GET_SB_BLOCKSIZE;
	if (blocksize != BLOCK_SIZE)
	{
		sb->s_dev = 0;
		UNLOCKSB;
		brelse(bh);
		printk("tifs: only %d block size supported (fs is %d)\n",
			BLOCK_SIZE, blocksize);
		MOD_DEC_USE_COUNT;
		return NULL;
	}

	/* Now it's loaded, we look to see how many directory entries there
	   are in it. Then, we load the whole thing. */
 
	superblocksize = sizeof(TIFS_Superblock) + sizeof(TIFS_Fileinfo)
		* GET_SB_NUMENTRIES;

	printk("tifs: superblock = %d bytes, %d entries, %d blocks\n", superblocksize, 
		GET_SB_NUMENTRIES, BLOCKS(superblocksize));

	/* Reload the superblock. */

	brelse(bh);

	ts = kmalloc(BLOCKS(superblocksize)*BLOCK_SIZE+BLOCK_SIZE, GFP_KERNEL);
	for(i=0; i<BLOCKS(superblocksize); i++)
	{
		bh = bread(dev, 1+i, BLOCK_SIZE);
		if (!bh)
		{
			sb->s_dev = 0;
			UNLOCKSB;
			kfree(ts);
			printk("tifs: unable to read extended superblock\n");
			MOD_DEC_USE_COUNT;
			return NULL;
		}

		memcpy(ts + (i*BLOCK_SIZE), bh->b_data, BLOCK_SIZE);

		brelse(bh);
	}

	/* We aren't an `offical' file system, so we don't have the convenience
	   of a superblock structure in struct super_block. We have to use a
	   pointer instead, which means we have to allocate our structure
	   dynamically. This is fine, though, seeing as it's so large. */

	sb->u.generic_sbp = ts;

	/* Init the system superblock structure. */

	sb->s_blocksize = blocksize;
	sb->s_blocksize_bits = BLOCK_SIZE_BITS;
	sb->s_magic = 0;
	sb->s_dev = dev;
	sb->s_op = &sops_block;

	/* Read the root directory's inode --- doesn't do much, but it keeps
	   the kernel happy. */

	UNLOCKSB;
	
	sb->s_mounted = iget(sb, TIFS_ROOT_INO);
	
	/* This shouldn't happen. */

	if (!sb->s_mounted)
	{
		sb->s_dev = 0;
		kfree(ts);
		printk("tifs: get root inode failed --- this is impossible!\n");
		MOD_DEC_USE_COUNT;
		return NULL;
	}

	DEBUG(("tifs: Ready.\n"));

	/* We're ready. */

	return sb;
}

/* I think this function's supposed to flush the superblock to disc. */

void tifs_write_super(struct super_block *sb)
{
	TIFS_Superblock *ts = sb->u.generic_sbp;
	int i;
	int dev = sb->s_dev;
	int superblocksize = sizeof(TIFS_Superblock) + sizeof(TIFS_Fileinfo)
		* GET_SB_NUMENTRIES;
	struct buffer_head *bh;

	for(i=0; i<BLOCKS(superblocksize); i++)
	{
		bh = bread(dev, 1+i, BLOCK_SIZE);
		if (!bh)
		{
			printk("tifs: unable to write extended superblock\n");
			return;
		}

		memcpy(bh->b_data, ts + (i*BLOCK_SIZE), BLOCK_SIZE);
		mark_buffer_dirty(bh, 1);

		brelse(bh);
	}
	
	sb->s_dirt = 0;
}

/* Write superblock and shut down, I think. */

void tifs_close_super(struct super_block *sb)
{
	TIFS_Superblock *ts = sb->u.generic_sbp;

	TRACE("tifs_close_super");

	LOCKSB;
	tifs_write_super(sb);
	TRACE("freeing extended superblock");
	kfree(ts);

	printk("tifs: File system on dev %s unmounted\n", 
		kdevname(sb->s_dev));

	sb->s_dev = 0;
	UNLOCKSB;
	MOD_DEC_USE_COUNT;
}

/* Get file system info. */

void tifs_statfs(struct super_block *sb, struct statfs *buf, int bufsize)
{
	TIFS_Superblock *ts = (TIFS_Superblock *)sb->u.generic_sbp;
	struct statfs tmp;

	tmp.f_type = sb->s_magic;
	tmp.f_bsize = GET_SB_BLOCKSIZE;
	tmp.f_blocks = GET_SB_DISCSIZE;
	tmp.f_bfree = GET_SB_DISCSIZE - GET_SB_FIRSTUSABLE - GET_SB_NUMBLOCKSUSED;
	tmp.f_bavail = tmp.f_bfree;
	tmp.f_files = GET_SB_NUMFILESUSED;
	tmp.f_ffree = GET_SB_NUMENTRIES - GET_SB_NUMFILESUSED;
	tmp.f_namelen = TIFS_NAMELENGTH;

	memcpy_tofs(buf, &tmp, bufsize);
}

/* --- Directory utilities ------------------------------------------------- */

/* Find an empty directory entry */

static int findslot(TIFS_Superblock *ts)
{
	int i;

	for(i=1; i<GET_SB_NUMENTRIES; i++)
		if (GET_FE_FILETYPE(i) == TIFS_EMPTY)
			return i;
	
	return -1;
}

/* Find the ino of a file given the name and the ino of the directory. */

static int findfile(TIFS_Superblock *ts, int dir, const char *name, int len)
{
	int i;
	char buf[TIFS_NAMELENGTH + 1];

	for (i=1; i<GET_SB_NUMENTRIES; i++)
		if ((GET_FE_FILETYPE(i) != TIFS_EMPTY) &&
		    (GET_FE_DIRID(i) == dir))
		{
			GET_FE_FILENAME(i, buf);
			if (strncmp(buf, name, len) == 0)
				return i;
		}
	
	return -1;
}

/* Checks a filename to see if it's valid. */

static int validatefilename(const char *name, int len)
{
	if (strlen(name) > TIFS_NAMELENGTH)
		return -ENAMETOOLONG;
	return 0;
}

/* --- Directory handling code --------------------------------------------- */

/* Read a directory. */

/* filp->f_pos is the offset into the directory. This is a completely opaque
   value (the first read has it set to 0). We use it to keep track of what
   inode we're at. 0 is ., 1 is .. (since ino 1 is the root dir, which
   isn't in a directory). The value is post-incremented. */

int tifs_readdir(struct inode *dir, struct file *filp, void *dirent,
	filldir_t filldir)
{
	INIT_SB_TS("tifs_readdir", dir);

	/* Sanity check. */

	if (!S_ISDIR(dir->i_mode))
		return -EBADF;

	if (dir->i_ino > GET_SB_NUMENTRIES)
		return -EBADF;

	switch((int)filp->f_pos)
	{
		/* Add . */

		case 0:		if (filldir(dirent, ".", 1, filp->f_pos, dir->i_ino) < 0)
					return 0;
				filp->f_pos++;
				break;
		
		/* Add .. */

		case 1:		{ 
					long parent = GET_FE_DIRID(dir->i_ino);
					if (filldir(dirent, "..", 2, filp->f_pos, parent) < 0)
						return 0;
					filp->f_pos++;
					break;
				}

		/* Normal files. */

		default:	{
					int index = filp->f_pos;

					while ((index<GET_SB_NUMENTRIES) &&
					    ((GET_FE_FILETYPE(index) == TIFS_EMPTY) ||
					     (GET_FE_DIRID(index) != dir->i_ino)))
						index++;

					if (index<GET_SB_NUMENTRIES)
					{
						char name[TIFS_NAMELENGTH+1];
						int namelen;

						GET_FE_FILENAME(index, name);
						namelen = strnlen(name, TIFS_NAMELENGTH);
						if (filldir(dirent, name, namelen, index, index) < 0)
							return 0;
					}

					if (index > GET_SB_NUMENTRIES)
						index = GET_SB_NUMENTRIES;
						
					filp->f_pos = index+1;
				}
	}

	/* Done. */

	dir->i_mtime = CURRENT_TIME;
	dir->i_dirt = 1;

	return 0;
}

/* Create a new file, type TIFS_FILE, but with no space allocated. */

static int create_raw(struct inode *dir, const char *name, int len, int mode,
	struct inode **rinode)
{
	int ino;
	int result;
	struct inode *inode;
	INIT_SB_TS("create_raw", dir);

	TRACE("create_raw");

	result = validatefilename(name, len);
	if (result)
		goto createraw_done;

	ino = findslot(ts);

	result = -ENOSPC;
	if (ino == -1)
		goto createraw_done;

	inode = get_empty_inode();
	inode->i_sb = dir->i_sb;
	inode->i_flags = dir->i_sb->s_flags;
	inode->i_count = 1;
	inode->i_nlink = 2;
	inode->i_dev = sb->s_dev;
	inode->i_uid = current->fsuid;
	inode->i_gid = (dir->i_mode & S_ISGID) ? dir->i_gid : current->fsgid;
	inode->i_dirt = 1;
	inode->i_ino = ino;
	inode->i_mtime = inode->i_atime = inode->i_ctime = CURRENT_TIME;
	inode->i_op = &iops_block;
	inode->i_mode = mode ;
	inode->i_blocks = inode->i_blksize = 0;
	insert_inode_hash(inode);

	/* Fill in dir entry. */

	SET_FE_START(ino, 0);
	SET_FE_BLOCKS(ino, 0);
	SET_FE_FILENAME(ino, name, len);
	SET_FE_FILETYPE(ino, TIFS_FILE);
	SET_FE_DIRID(ino, dir->i_ino);
	SET_FE_MODE(ino, mode & 07777);
	SET_FE_SIZE(ino, 0);

	/* Update superblock stats. */

	SET_SB_NUMFILESUSED(GET_SB_NUMFILESUSED+1);
	
	/* Touch dir */
	
	dir->i_mtime = dir->i_atime = CURRENT_TIME;
	dir->i_dirt = 1;

	*rinode = inode;
	result = 0;  	
createraw_done:
	return result;
}

/* Creates an ordinary file. */

int tifs_create(struct inode *dir, const char *name, int len, int mode,
	struct inode **rinode)
{
	int ino;
	int result;
	int firstblock;
	int i;
	struct inode *inode;
	INIT_SB_TS("tifs_create", dir);

	TRACE("tifs_create");

	/* Create a file */

	LOCKSB;

	result = validatefilename(name, len);
	if (result)
		goto create_done;

	result = create_raw(dir, name, len, mode, rinode);
	if (result)
		goto create_done;

	inode = *rinode;
	ino = inode->i_ino;

	/* Scan for a hole in the disc to put the file in. */

	firstblock = GET_SB_FIRSTUSABLE;

	do {
		DEBUG(("tifs: scanning for free block; @%d\n", firstblock));

		result = 0;
		for (i=2; i<GET_SB_NUMENTRIES; i++)
			if ((GET_FE_BLOCKS(i) > 0) &&
			    (GET_FE_START(i) == firstblock))
			{
				result = 1;
				firstblock += GET_FE_BLOCKS(i);
			}
	} while ((firstblock < GET_SB_DISCSIZE) && result);

	DEBUG(("tifs: got hole, %d\n", firstblock));

	result = -ENOSPC;
	if (firstblock >= GET_SB_DISCSIZE)
	{
		iput(inode);
		goto create_done;
	}

	/* Fill in dir entry. */

	SET_FE_START(ino, firstblock);
	SET_FE_BLOCKS(ino, 1);
	SET_FE_SIZE(ino, 0);

	SET_SB_NUMBLOCKSUSED(GET_SB_NUMBLOCKSUSED+1);
	
	result = 0;  	
create_done:
	UNLOCKSB;
	iput(dir);
	return result;
}

/* Creates a special file of some type. */

int tifs_mknod(struct inode *dir, const char *name, int len, int mode,
	int rdev)
{
	int ino;
	int result;
	struct inode *inode;
	INIT_SB_TS("tifs_mknod", dir);

	TRACE("tifs_mknod");

	result = validatefilename(name, len);
	if (result)
	{
		iput(dir);
		return result;
	}

	/* Create the file */

	LOCKSB;

	result = create_raw(dir, name, len, mode, &inode);
	if (result)
	{
		iput(dir);
		UNLOCKSB;
		return result;
	}

	ino = inode->i_ino;

	/* Fill in the inode special fields. */

        if (S_ISCHR(mode))
        {
        	SET_FE_FILETYPE(ino, TIFS_CHRDEV);
                inode->i_op = &chrdev_inode_operations;
        }
        else if (S_ISBLK(mode))
        {
        	SET_FE_FILETYPE(ino, TIFS_BLKDEV);
                inode->i_op = &blkdev_inode_operations;
        }
        else if (S_ISFIFO(mode))
        {
        	SET_FE_FILETYPE(ino, TIFS_FIFO);
                init_fifo(inode);
        }
        else if (S_ISSOCK(mode))
        	SET_FE_FILETYPE(ino, TIFS_SOCKET);
        else
        {
        	printk("tifs: mknod called with mode %o!\n", mode);
        	SET_FE_FILETYPE(ino, TIFS_CHRDEV);
        	mode &= 07777;
        	mode |= S_IFCHR;
        	inode->i_mode = mode;
        	rdev = 0;
        }

        if (S_ISBLK(mode) || S_ISCHR(mode))
                inode->i_rdev = to_kdev_t(rdev);

	/* Fill in dir entry. */

	SET_FE_SIZE(ino, rdev);

	result = 0;  	
/* mknod_done: */
	UNLOCKSB;
	iput(inode);
	iput(dir);
	return result;
}
/* Looks up a name in a directory. */

int tifs_lookup(struct inode *dir, const char *name, int len,
	struct inode **rinode)
{
	int ino;
	int result;
	INIT_SB_TS("tifs_lookup", dir);

	/* Sanity check. */

	result = validatefilename(name, len);
	if (result)
	{
		iput(dir);
		return result;
	}

	if (!dir)
		return -ENOENT;

	if (!S_ISDIR(dir->i_mode))
	{
		iput(dir);
		return -ENOENT;
	}

	*rinode = NULL;

	/* Check for . and .. (not in the TiFS directory) */

	if ((len == 0) || ((len==1) && (name[0] == '.')))
	{
		*rinode = dir;
		return 0;
	}

	if ((len==2) && (name[0] == '.') && (name[1] == '.'))
	{
		int parent = GET_FE_DIRID(dir->i_ino);

		*rinode = iget(dir->i_sb, parent);
		iput(dir);
		return (*rinode)?0:-EACCES;
	}

	/* Scan the directory, looking for the file. */

	ino = findfile(ts, dir->i_ino, name, len);

	if (ino == -1)
	{
		iput(dir);
		return -ENOENT;
	}

	*rinode = iget(dir->i_sb, ino);

	dir->i_mtime = CURRENT_TIME;
	dir->i_dirt = 1;

	iput(dir);
	return 0;
}

/* Delete a file. */

int tifs_unlink(struct inode *dir, const char *name, int len)
{
	int result = 0;
	int ino;
	INIT_SB_TS("tifs_unlink", dir);

	TRACE("tifs_unlink");

	/* Get the inode of the file to delete. */

	LOCKSB;

	result = validatefilename(name, len);
	if (result)
		goto unlink_done;

	ino = findfile(ts, dir->i_ino, name, len);

	if (ino == -1)
	{
		result = -ENOENT;
		goto unlink_done;
	}

	/* If the file is a directory, disallow unlinking. Must use rmdir
	   instead. */

	if (GET_FE_FILETYPE(ino) == TIFS_DIR)
	{
		result = -EPERM;
		goto unlink_done;
	}

	/* Nuke the file. */

	SET_FE_FILETYPE(ino, TIFS_EMPTY);
	SET_FE_BLOCKS(ino, 0);
	result = 0;

	dir->i_mtime = dir->i_ctime = CURRENT_TIME;
	dir->i_dirt = 1;

unlink_done:
	UNLOCKSB;
	iput(dir);
	return result;
}

/* Make a directory. */

int tifs_mkdir(struct inode *dir, const char *name, int len, int mode)
{
	int result = 0;
	int ino;
	INIT_SB_TS("tifs_mkdir", dir);

	/* Check to see if the directory exists. */

	LOCKSB;

	result = validatefilename(name, len);
	if (result)
		goto mkdir_done;
		
	result = -EEXIST;
	if (findfile(ts, dir->i_ino, name, len) != -1)
		goto mkdir_done;

	/* Create the directory. */

	ino = findslot(ts);

	result = -ENOSPC;
	if (ino<0)
		goto mkdir_done;

	SET_SB_NUMFILESUSED(GET_SB_NUMFILESUSED + 1);

	SET_FE_FILENAME(ino, name, len);
	SET_FE_FILETYPE(ino, TIFS_DIR);
	SET_FE_DIRID(ino, dir->i_ino);
	SET_FE_START(ino, 0);
	SET_FE_MODE(ino, mode & 07777);
	SET_FE_SIZE(ino, 0);

	result = 0;

	dir->i_mtime = dir->i_ctime = CURRENT_TIME;
	dir->i_dirt = 1;

mkdir_done:
	iput(dir);
	UNLOCKSB;
	return result;
}

/* Remove a directory. */

int tifs_rmdir(struct inode *dir, const char *name, int len)
{
	int result = 0;
	int ino;
	int i;
	INIT_SB_TS("tifs_rmdir", dir);

	/* Check to see if the directory exists. */

	LOCKSB;
	result = validatefilename(name, len);
	if (result)
		goto rmdir_done;
		
	ino = findfile(ts, dir->i_ino, name, len);

	result = -ENOENT;
	if (ino == -1)
		goto rmdir_done;

	/* Make sure that it really *is* a directory. */

	result = -ENOTDIR;
	if (GET_FE_FILETYPE(ino) != TIFS_DIR)
		goto rmdir_done;

	/* Make sure that the directory is empty. */

	result = -EBUSY;
	for (i=1; i<GET_SB_NUMENTRIES; i++)
		if ((GET_FE_FILETYPE(i) != TIFS_EMPTY) &&
		    (GET_FE_DIRID(i) == ino))
			goto rmdir_done;
	
	/* Do the removal. */

	result = 0;
	SET_FE_FILETYPE(ino, TIFS_EMPTY);

	dir->i_mtime = dir->i_ctime = CURRENT_TIME;
	dir->i_dirt = 1;

rmdir_done:
	UNLOCKSB;
	iput(dir);
	return result;
}

/* Rename a file */

int tifs_rename(struct inode *olddir, const char *oldname, int oldlen,
		struct inode *newdir, const char *newname, int newlen,
		int mustbedir)
{
	int result = 0;
	int ino;
	INIT_SB_TS("tifs_rename", olddir);

	TRACE("tifs_rename");
	
	/* Check to see if the source file exists. */

	LOCKSB;

	result = validatefilename(oldname, oldlen);
	if (result)
		goto rename_done;
	result = validatefilename(newname, newlen);
	if (result)
		goto rename_done;
		
	ino = findfile(ts, olddir->i_ino, oldname, oldlen);

	result = -ENOENT;
	if (ino == -1)
		goto rename_done;
	
	/* Check to see if the destination file *doesn't* exist. */

	result = -EEXIST;
	if (findfile(ts, newdir->i_ino, newname, newlen) != -1)
		goto rename_done;

	/* Do the renaming. */

	SET_FE_FILENAME(ino, newname, newlen);
	SET_FE_DIRID(ino, newdir->i_ino);
	result = 0;

	olddir->i_mtime = olddir->i_ctime = CURRENT_TIME;
	olddir->i_dirt = 1;
	newdir->i_mtime = newdir->i_ctime = CURRENT_TIME;
	newdir->i_dirt = 1;

rename_done:
	UNLOCKSB;
	iput(olddir);
	iput(newdir);
	return result;
}

/* --- File operations ----------------------------------------------------- */

/* bmap. This should return the absolute block number given the file and the
   relative block number. The kernel will then go away and do its own thing
   to read the file. */

int tifs_bmap(struct inode *inode, int block)
{
	INIT_SB_TS("tifs_bmap", inode);

	return GET_FE_START(inode->i_ino) + block;
}

/* Truncates the file on disc to the length stored in the inode. */

void tifs_truncate(struct inode *inode)
{
	INIT_SB_TS("tifs_truncate", inode);

        if (!(S_ISREG(inode->i_mode) ||
	     S_ISDIR(inode->i_mode) ||
             S_ISLNK(inode->i_mode)))
		return;

	inode->i_mtime = inode->i_ctime = CURRENT_TIME;
	inode->i_dirt = 1;
}

/* This is the biggie: writes data to disc from user space. Also extends the
   file, if necessary, if possible. */

int tifs_file_write(struct inode *inode, struct file *filp,
		    const char *buffer, int len)
{
	int ino;
	int pos;
	int written;
	int c;
	char *p;
	struct buffer_head *bh;
	INIT_SB_TS("tifs_file_write", inode);

	ino = inode->i_ino;

	if (filp->f_flags & O_APPEND)
		pos = inode->i_size;
	else
		pos = filp->f_pos;

	if (BLOCKS(pos+len) > GET_FE_BLOCKS(ino))
	{
		/* Need to extend the file. */

		int holesize = GET_SB_DISCSIZE - GET_FE_START(ino);
		int i;

		LOCKSB;

		DEBUG(("tifs: Extending %d. pos=%d (%d)\n", ino,
			pos, BLOCKS(pos+len)));

		for (i=1; i<GET_SB_NUMENTRIES; i++)
			if ((GET_FE_FILETYPE(i) == TIFS_FILE) &&
				(GET_FE_START(i) > GET_FE_START(ino)) &&
				(GET_FE_START(i) - GET_FE_START(ino) < holesize))
					holesize = GET_FE_START(i) - GET_FE_START(ino);

		DEBUG(("tifs: Extending %d: current size=%d, can extend to %d\n",
			ino, GET_FE_BLOCKS(ino), holesize));

		if (BLOCKS(pos+len) > holesize)
		{
			char buf[TIFS_NAMELENGTH+1];

			/* Can't do a simple extend. */

			GET_FE_FILENAME(ino, buf);
			printk("tifs: Can't extend ino %d (%d/%s)\n",
				ino, GET_FE_DIRID(ino), buf);
			UNLOCKSB;
			return -ENOSPC;
		}

		/* We *can* do a simple extend, so do it. */

		SET_FE_BLOCKS(ino, BLOCKS(pos+len));

		UNLOCKSB;
	}

	/* Write data. */

	written = 0;
	while (written < len)
	{
		/* Open block */

		DEBUG(("tifs: Writing, pos=%d\n", pos));

		bh = bread(sb->s_dev, tifs_bmap(inode, pos/BLOCK_SIZE),
			BLOCK_SIZE);
		if (!bh)
		{
			if (!written)
				return -ENOSPC;
			break;
		}

		/* Calculate how much to write into this block. */

		c = BLOCK_SIZE - (pos % BLOCK_SIZE);
		if (c > len - written)
			c = len - written;
		
		/* Make sure our buffer's up-to-date. */

		if ((c != BLOCK_SIZE) && !buffer_uptodate(bh))
		{
			ll_rw_block(READ, 1, &bh);
			wait_on_buffer(bh);
			if (!buffer_uptodate(bh))
			{
				/* Block busy. Can't write. */

				brelse(bh);
				if (!written)
					written = -EIO;
				break;
			}
		}

		/* Tell the kernel to copy the data onto disc when it gets
		   the chance. */

		p = (pos % BLOCK_SIZE) + bh->b_data;
		memcpy_fromfs(p, buffer, c);
		update_vm_cache(inode, pos, p, c);
		mark_buffer_uptodate(bh, 1);
		mark_buffer_dirty(bh, 0);
		brelse(bh);
		pos += c;
		written += c;
		buffer += c;
	}

	/* Update inode, etc. */

	if (pos > inode->i_size)
		inode->i_size = pos;
	inode->i_mtime = inode->i_atime = CURRENT_TIME;
	filp->f_pos = pos;
	inode->i_dirt = 1;

	DEBUG(("tifs: written %d\n", written));

	return written;
}

/* --- Inode operations ---------------------------------------------------- */

/* Reads an inode given the ino. */

void tifs_read_inode(struct inode *inode)
{
	int ino;
	INIT_SB_TS("tifs_read_inode", inode);

	ino = inode->i_ino;
	inode->i_op = NULL;
	inode->i_mode = 0;

	if (ino >= GET_SB_NUMENTRIES)
	{
		printk("tifs: Bad inode number on dev %s: %d is out of range\n",
			kdevname(inode->i_dev), ino);
		return;
	}

        inode->i_mode = GET_FE_MODE(ino) & 07777;
        inode->i_uid = GET_FE_UID(ino);
        inode->i_gid = GET_FE_GID(ino);
        inode->i_nlink = 1;
        inode->i_size = GET_FE_SIZE(ino);
        inode->i_mtime = GET_FE_MTIME(ino);
        inode->i_atime = GET_FE_ATIME(ino);
        inode->i_ctime = GET_FE_CTIME(ino);
        inode->i_blocks = GET_FE_BLOCKS(ino);
        inode->i_blksize = GET_SB_BLOCKSIZE;

        switch(GET_FE_FILETYPE(ino))
        {
        	case TIFS_FILE:		inode->i_op = &iops_block;
					inode->i_mode |= S_IFREG;
					break;
		
		case TIFS_DIR:		inode->i_op = &iops_dir_block;
					inode->i_mode |= S_IFDIR;
					break;

		case TIFS_LINK:		inode->i_op = &iops_symlink_block;
					inode->i_mode |= S_IFLNK;
					break;

		case TIFS_CHRDEV:	inode->i_op = &chrdev_inode_operations;
					inode->i_mode |= S_IFCHR;
					inode->i_rdev = to_kdev_t(GET_FE_SIZE(ino));
					inode->i_size = 0;
					break;
		
		case TIFS_BLKDEV:	inode->i_op = &blkdev_inode_operations;
					inode->i_mode |= S_IFBLK;
					inode->i_rdev = to_kdev_t(GET_FE_SIZE(ino));
					inode->i_size = 0;
					break;
		
		case TIFS_FIFO:		inode->i_mode |= S_IFIFO;
					init_fifo(inode);
					break;

		case TIFS_SOCKET:	inode->i_mode |= S_IFSOCK;
					break;
		
		default:		printk("tifs: inconsistent FS "
					  "(don't know what type %d is)\n",
					  GET_FE_FILETYPE(ino));
					inode->i_op = &iops_block;
					inode->i_mode |= S_IFREG;
					break;
	}
}

/* Write an inode. */

void tifs_write_inode(struct inode *inode)
{
	int ino;
	INIT_SB_TS("tifs_write_inode", inode);

	ino = inode->i_ino;

	if ((ino == 0) || (ino >= GET_SB_NUMENTRIES))
	{
		printk("Bad inode number on dev %s: %d is out of range\n",
			kdevname(inode->i_dev), ino);
		inode->i_dirt = 0;
		return;
	}

	LOCKSB;
	SET_FE_MODE(ino, inode->i_mode & 07777);
	SET_FE_UID(ino, inode->i_uid);
	SET_FE_GID(ino, inode->i_gid);
	SET_FE_SIZE(ino, inode->i_size);
	SET_FE_MTIME(ino, inode->i_mtime);
	SET_FE_ATIME(ino, inode->i_atime);
	SET_FE_CTIME(ino, inode->i_ctime);

	if (S_ISCHR(inode->i_mode) || S_ISBLK(inode->i_mode))
		SET_FE_SIZE(ino, kdev_t_to_nr(inode->i_rdev));

	inode->i_dirt = 0;
	UNLOCKSB;
}

/* Close an inode. */

void tifs_close_inode(struct inode *inode)
{
	INIT_SB_TS("tifs_close_inode", inode);

	if (inode->i_nlink)
		return;

	LOCKSB;
	inode->i_size = 0;
	UNLOCKSB;
	tifs_truncate(inode);
	clear_inode(inode);
}

/* --- Symlink ------------------------------------------------------------- */

/* Creates a symlink */

int tifs_symlink(struct inode *dir, const char *name, int len,
	const char *symname)
{
	struct inode *inode;
	int result;
	int ino;
	struct buffer_head *bh;
	INIT_SB_TS("tifs_symlink", dir);

	TRACE("tifs_create");

	/* Create a file */

	result = validatefilename(name, len);
	if (result)
	{
		iput(dir);
		return result;
	}
		
	result = tifs_create(dir, name, len, S_IFLNK | 0777, &inode);
	iput(dir);
	if (result)
		return result;
	
	/* Turn it into a symlink */

	ino = inode->i_ino;
	iput(inode);

	LOCKSB;

	SET_FE_FILETYPE(ino, TIFS_LINK);

	/* When the file was created, one block was allocated. We
	   put the destination path into this block. */

	bh = bread(sb->s_dev, tifs_bmap(inode, 0), BLOCK_SIZE);
	if (!bh)
	{
		UNLOCKSB;
		return -EIO;
	}
	
	strncpy(bh->b_data, symname, BLOCK_SIZE - 1);
	
	mark_buffer_dirty(bh, 1);
	brelse(bh);

	UNLOCKSB;
	return 0;
}

/* Follows a symlink */

int tifs_followlink(struct inode *dir, struct inode *inode,
        int flag, int mode, struct inode **rinode)
{
        int result;
        struct buffer_head *bh;
	INIT_SB_TS("tifs_followlink", dir);

	TRACE("tifs_followlink");

        *rinode = NULL;

        /* Increase the links followed count. */

        if (!dir) {
                dir = current->fs->root;
                dir->i_count++;
        }

        /* Sanity check */

        if (!inode) {
                iput(dir);
                return -ENOENT;
        }

        /* Reached the end? */

        if (!S_ISLNK(inode->i_mode)) {
                iput(dir);
                *rinode = inode;
                return 0;
        }

        /* Too many links? */

        if (current->link_count > 5) {
                iput(inode);
                iput(dir);
                return -ELOOP;
        }

	/* Read the destination path. */

	bh = bread(sb->s_dev, tifs_bmap(inode, 0), BLOCK_SIZE);
	if (!bh)
	{
		iput(inode);
		iput(dir);
		return -EIO;
	}

	/* ...and recurse. */

        iput(inode);
        current->link_count++;
        result = open_namei(bh->b_data, flag, mode, rinode, dir);
        current->link_count--;
        brelse(bh);
        return result;
}

/* Read a symlink name. */

int tifs_readlink(struct inode *inode, char *buffer, int buflen)
{
        struct buffer_head *bh;
        int i;
        char c;
	INIT_SB_TS("tifs_readlink", inode);

	TRACE("tifs_readlink");

	/* Check to make sure that this really is a link. */

        if (!S_ISLNK(inode->i_mode)) {
                iput(inode);
                return -EINVAL;
        }

        /* Read the link. */

        if (buflen > 1023)
                buflen = 1023;

	bh = bread(sb->s_dev, tifs_bmap(inode, 0), BLOCK_SIZE);
        iput(inode);
        if (!bh)
                return -EIO;

	/* Copy the link name into user space. */

        i = 0;
        while (i<buflen && (c = bh->b_data[i])) {
                i++;
                put_user(c,buffer++);
        }

        brelse(bh);
        return i;
}

