/* tifsck.c
 * Checks out a TiFS file system.
 * (C) 1996 David Given
 */

#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>
#include <string.h>
#include <linux/fs.h>
#include "tifs.h"

#define BLOCKS(size) \
	(((size) / BLOCK_SIZE) + 1)

void help(void);
void action_view(void);

static struct option options[] =
{
	{"device",	1,	NULL,	'd'},
	{"view",	0,	NULL,	'v'},
	{"help",	0,	NULL,	'h'},
	{NULL,		0,	NULL,	0},
};

enum {
	ACTION_NONE,
	ACTION_VIEW
} action = ACTION_NONE;

char *device = NULL;
int superblocksize;
TIFS_Superblock *ts;

int main(int argc, char **argv)
{
	int i;
	FILE *f;

	while(1)
	{
		int index = 0;

		i = getopt_long(argc, argv,
			"hvd:", options, &index);
		if (i == -1)
			break;

		switch (i)
		{
			case 'd':	device = strdup(optarg);
					break;

			case 'v':	action = ACTION_VIEW;
					break;

			case 'h':	help();
					exit(0);
					break;

			default:	help();
					exit(1);
		}
	}

	if (device == NULL)
	{
		fprintf(stderr, "tifsck: You must specify a device (-h for help)\n");
		exit(1);
	}

	/* Load the superblock */

	f = fopen(device, "r+");

	ts = calloc(sizeof(TIFS_Superblock), 1);

	fseek(f, BLOCK_SIZE, SEEK_SET);
	fread(ts, sizeof(TIFS_Superblock), 1, f);

	if ((ts->magic[0] != 'T') ||
		(ts->magic[1] != 'i') ||
		(ts->magic[2] != 'F') ||
		(ts->magic[3] != 'S'))
	{
		fclose(f);
		fprintf(stderr, "Magic number wrong --- is this a TiFS file system?\n");
		exit(1);
	}

	superblocksize = sizeof(TIFS_Superblock) + sizeof(TIFS_Fileinfo)
		* ts->numentries;

	ts = realloc(ts, superblocksize);
	fseek(f, BLOCK_SIZE, SEEK_SET);
	fread(ts, superblocksize, 1, f);

	fclose(f);

	switch(action)
	{
		case ACTION_NONE:	fprintf(stderr, "You must specify an action\n");
					exit(1);

		case ACTION_VIEW:	action_view();
					break;
	}

	exit(0);
}

void action_view(void)
{
	int i;
	TIFS_Fileinfo *f;
	const char *filetype[] =
	{
		"EMPTY",
		"FILE",
		"LINK",
		"DIR",
		"BLKDEV",
		"CHRDEV",
		"FIFO"
	};

	printf(
"TiFS superblock:\n"
"    Block size = %d bytes\n"
"    Disc size = %d blocks (%d kB)\n"
"    First usable block = %d\n"
"    Directory size = %d entries (%d used, %d free), %d blocks (%d kB)\n"
"    Space used = %d blocks (%d kB)\n"
"    Boot image @%05d+%05d\n",
	ts->blocksize,
	ts->discsize, (ts->discsize * ts->blocksize)/0x400,
	ts->firstusable,
	ts->numentries, ts->numfilesused, ts->numentries - ts->numfilesused,
	BLOCKS(superblocksize), (BLOCKS(superblocksize) * ts->blocksize)/0x400,
	ts->numblocksused, (ts->numblocksused * ts->blocksize)/0x400,
	ts->boot_start, ts->boot_size);

	printf("\nDirectory:\n");

	for(i=1; i<ts->numentries; i++)
	{
		f = &ts->file[i];
		if (f->filetype == TIFS_EMPTY)
			continue;
		
		printf("%3d (%3d)\t%s\t@%05d+%05d   %4o %5d %5d\t%d\t%s\n",
			i, f->dirid, filetype[f->filetype], f->start, f->blocks,
			f->mode, f->uid, f->gid, f->size, f->filename);
	}
}

void help(void)
{
	printf(
"mktifs [options]\n"
"    -d --device       Which device to process\n"
"\n"
"Actions:\n"
"    -v --view         Displays info on the file system\n"
"    -h --help         Displays this message\n");
}

